#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#       bvpy.domains.abstract
#
#       File author(s):
#           Florian Gacon <florian.gacon@inria.fr>
#
#       File contributor(s):
#           Florian Gacon <florian.gacon@inria.fr>
#           Olivier Ali <olivier.ali@inria.fr>
#
#       File maintainer(s):
#           Olivier Ali <olivier.ali@inria.fr>
#
#       Copyright © by Inria
#       Distributed under the LGPL License..
#       See accompanying file LICENSE.txt or copy at
#           https://www.gnu.org/licenses/lgpl-3.0.en.html
#
# -----------------------------------------------------------------------
import time

import gmsh
import sys
import os
import shutil
from bvpy import logger as log
from bvpy import MPI_COMM_WORLD
import meshio as mio
from abc import ABC, abstractmethod
import tempfile

import fenics as fe
import numpy as np

# Available meshing algorithm from gmsh (experimental algorithms are removed)
# See https://gmsh.info/doc/texinfo/gmsh.html#Choosing-the-right-unstructured-algorithm
# to have more information about what algo fits your need. Briefy :
# - complex curved surfaces (2D) --> MeshAdapt
# - high element quality (2D/3D) --> Frontal-Delaunay
# - very large meshes --> Delaunay (2D) / HXT (3D)
AVAILABLE_MESHING_ALGO_2D = {'MeshAdapt': 1, 'Automatic': 2, 'Delaunay': 5, 'Frontal-Delaunay': 6}
AVAILABLE_MESHING_ALGO_3D = {'Delaunay': 1, 'Frontal-Delaunay': 4, 'HXT': 10}

# Available cell type for the mesh (hybrid mesh are not allowed)
AVAILABLE_CELL_TYPE = ['tetra', 'triangle', 'line']
CELL_TYPE = {'tetra': 3, 'triangle': 2, 'line': 1, 'vertex': 0} # cell_type : topological_dimension

def _generate_default_boundary_data(mesh: fe.Mesh) -> fe.MeshFunction:
    """Sets the boundary of a mesh.

    Parameters
    ----------
    mesh : :class:`Mesh<Dolfin.cpp.mesh.Mesh>`
        The mesh on which we want to define boundaries.

    Returns
    -------
    :class:`MeshFunction<Dolfin.cpp.mesh.MeshFunction>`
        A fenics MeshFunction that marks (with the index 0)
        the boundary of the considered mesh.

    """
    mf = fe.MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
    mf.set_all(sys.maxsize)
    subdomain = fe.CompiledSubDomain('on_boundary')
    subdomain.mark(mf, 0)
    return mf


def _generate_default_cell_data(mesh: fe.Mesh) -> fe.MeshFunction:
    """Sets the body (i.e. inner part) of a mesh.

    Parameters
    ----------
    mesh : :class:`Mesh<Dolfin.cpp.mesh.Mesh>`
        The mesh on which we want to mark the body.

    Returns
    -------
    :class:`MeshFunction<Dolfin.cpp.mesh.MeshFunction>`
        A fenics MeshFunction that marks (with the index 0 ?)
        the boundary of the considered mesh.

    Notes
    -----
    #TODO -> is it normal that the body and the boundary
    of the mesh are marked with the same label ? Or this could mean that I did
    not understand the purpose of these functions.

    """
    mf = fe.MeshFunction("size_t", mesh, mesh.topology().dim())
    mf.set_all(0)
    return mf


class AbstractDomain(ABC):
    """The most general class that defines an integration domain.

    Attributes
    ----------
    cdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunctionSizet>`
        Description of attribute `cdata`.
    bdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunctionSizet>`
        Description of attribute `bdata`.
    dx : :class:`Measure<ufl.measure.Measure>`
        An integral measure within the domain of interest.
    ds : :class:`Measure<ufl.measure.Measure>`
        An integral measure on the border of the domain of interest.
    mesh : :class:`Mesh<dolfin.cpp.mesh.Mesh>`
        A mesh representation (topology and geometry) of the considered domain.
    model : :class:`Model<gmsh.model>`
        A purely geometrical description of the domain.
    domain_type : str
        Name of the geometrical model of the domain.
    details : str
        A printable description of the domain main characteristics.

    """
    _model = gmsh.model
    _domain_instances = 0

    def __init__(self,
                 cell_type: str = 'triangle',
                 cell_size=1,
                 verbosity: bool = False,
                 algorithm: str = 'Delaunay',
                 dim: int = 3,
                 clear: bool = False):
        """Generates an instance of the AbstractDomain class.

        Parameters
        ----------
        cell_type : str
            Type of element used to tile the domain.
            (the default is 'triangle').
        cell_size : float, [float, float]
            Characteristic size of the tiling elements (the default is 1). You can relax this constraint by defining
            cell_size as [min_cell_size, max_cell_size].
        verbosity : bool
            If `True`, log is printed (the default is False).
        algorithm : str
            Name of the algorithm used to compute the elements.
            (the default is 'Delaunay').

        Note
        ----
        The following cell types can be used to tile the considered domain:
            * 'line' (1D element)
            * 'triangle' (2D)
            * 'tetra' (3D).
            
        Accepted meshing algorithms depends on the cell_type dimension:
            * 2D ('triangle') : 'Delaunay', 'Frontal-Delaunay', 'MeshAdapt', 'Automatic'
            * 3D ('tetra') : 'Delaunay', 'Frontal-Delaunay', 'HXT'
        """

        super().__init__(cell_type=cell_type, cell_size=cell_size,
                         verbosity=verbosity, algorithm=algorithm)

        AbstractDomain._domain_instances += 1

        if AbstractDomain._domain_instances == 1:
            # - Initialize GMSH and the temp dir necessary to store the mesh
            gmsh.initialize() # required for all processes
            self.set_verbosity(verbosity)
            log.debug(f'Initialize GMSH for {MPI_COMM_WORLD.Get_rank()}')

            if MPI_COMM_WORLD.Get_rank() == 0:
                self._mesh_dir = tempfile.TemporaryDirectory()  # keep reference to avoid suppression of temp folder
                dir_name = self._mesh_dir.name
            else:
                dir_name = None

            # Broadcast to other processes
            self._dir_name = MPI_COMM_WORLD.bcast(dir_name, root=0)

            log.debug(f'TempFile {self._dir_name} for {MPI_COMM_WORLD.Get_rank()}')

        if clear:
            gmsh.clear()

        self._characteristic_size = None # TODO : do we need to init these values here ?
        self._cell_type = cell_type
        self._cell_size = cell_size
        self._algorithm = algorithm

        self.set_cell_type(cell_type)
        self.set_cell_size(cell_size)
        self.set_meshing_algorithm(algorithm)
        self.set_dim(dim)

        self.dim = dim # TODO : should it be an hidden prop?
        self.mesh = None
        self.cdata = None
        self.bdata = None
        self.sub_domain_names = None # specify explicit names for labels in cdata (MeshFunctionSizet)
        self.sub_boundary_names = None # specify explicit names for labels in bdata (MeshFunctionSizet)


        self.dx = fe.dx
        self.ds = fe.ds

        self.surfaces = dict()
        self.volumes = dict()

    def __del__(self):
        """Clears the temporary file containing the domain.

        """
        #self._mesh_dir.cleanup()

        if (MPI_COMM_WORLD.Get_rank() == 0) and hasattr(self, '_dir_name'):
            if os.path.isdir(self._dir_name):
                # Remove the temporary folder and its content using process of rank 0
                log.debug(f'Clear temp file for {MPI_COMM_WORLD.Get_rank()} at {self._dir_name} instances')
                shutil.rmtree(self._dir_name)


        AbstractDomain._domain_instances -= 1
        if AbstractDomain._domain_instances == 0:
            gmsh.finalize() # for all processes
            log.debug(
                f'Finalize GMSH  for {MPI_COMM_WORLD.Get_rank()}')

    def __repr__(self):
        """Returns a short description of the considered domain.

        Returns
        -------
        str
            A printable short expression of the domain.

        """
        return '<'+str(self.domain_type)+', at '+str(hex(id(self)))+'>'

    def __str__(self):
        """Returns a detailed description of the considered domain.

        Returns
        -------
        str
            A printable detailed description of the domain.

        """
        return self._details

    @property
    def domain_type(self):
        """str: Name of the geometrical model of the domain.

        """
        return self.__class__.__name__

    @property
    def model(self):
        """:class:`Model<gmsh.model>`: The (gmsh) geometrical model used
                                       to define the domain.

        """
        return AbstractDomain._model

    @property
    def _details(self):
        """str: A detailed printable description of the domain.

        """

        description = f'''Domain
------
    * Shape: {self.domain_type}
    * Dimension: {self.dim}\n'''

        for name, val in self._characteristic_size.items():
            description += f'    * {name}: {val:.2e}\n'

        if self.mesh is not None:
            description += '\n' + f'''Meshing
-------
        * Algorithm: {self._algorithm}
        * Cell type: {self._cell_type}
        * Number: {len(self.mesh.cells())}
        * Resolution (i.e. prescribed element size): {self._cell_size:.2e}
        * Actual size (min, max): ({self.mesh.hmin():.2e},\
        {self.mesh.hmax():.2e})
    '''
            cells_quality = np.array([c.radius_ratio()
                                      for c in fe.cells(self.mesh)])
            mn_c_qual = cells_quality.mean()
            std_c_qual = cells_quality.std()

            description += "    * Cells quality:"
            description += f" {mn_c_qual:.2e} +/- {std_c_qual:.2e} \n"

        else:
            description = "Mesh not generate yet."
            description += " Use self.discretize() to to so."

        return description

    def synchronize(self):
        """synchronizes the CAD model with the gmsh model.

        """
        self.factory.synchronize()

    @abstractmethod
    def geometry(self, *args, **kwargs):
        """Computes the geometry of the domain.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Note
        ----
        This abstract method must be defined within daughter class
        depending on the specific need.

        """
        pass

    def create_geometry(self, *args, **kwargs):
        """ Routine to control the update of the geometry if multi-threading is used"""

        if MPI_COMM_WORLD.Get_rank() == 0:
            log.debug('Create geometry of the domain!')
            # - only the process of rank 0 updates the geometry
            self.geometry(*args, **kwargs)
    def run_gui(self):
        """Generates a quick gmsh gui with the current model in it.

        """
        gmsh.fltk.run()

    def set_dim(self, dim: int):
        """Sets the dimension of the embedded space

        Parameters
        ----------
        dim : int
            Dimension of the embedded space ?

        """
        if (dim < CELL_TYPE[self._cell_type]) or (type(dim) is not int):
            raise ValueError(f'Dimension of the embedded space need to be equal or larger than the dimension'
                             f' of the finite element. Here, you are using "{self._cell_type}" meaning'
                             f' that dim >= {CELL_TYPE[self._cell_type]}.')

        self.dim = dim

    def set_cell_size(self, cell_size):
        """Sets the characteristic size of the domain elements.

        Parameters
        ----------
        cell_size : float, [float, float]
            The cell_size value we want to use.

        """
        if type(cell_size) in (float, int):
            gmsh.option.setNumber("Mesh.CharacteristicLengthMin", cell_size)
            gmsh.option.setNumber("Mesh.CharacteristicLengthMax", cell_size)
        elif ((type(cell_size) is list) and (len(cell_size) == 2)
              and all(type(ele) in (int, float) for ele in cell_size)):
            gmsh.option.setNumber("Mesh.CharacteristicLengthMin", cell_size[0])
            gmsh.option.setNumber("Mesh.CharacteristicLengthMax", cell_size[1])
        else:
            raise ValueError('cell_size can be a number or a list of two numbers')

        self._cell_size = cell_size
    def set_cell_type(self, cell_type: str):
        """Sets the type of finite element (triangle, tetra) you want to use for your mesh

        Parameters
        ----------
        cell_type: str
            The cell_type you want to use

        """
        if cell_type not in AVAILABLE_CELL_TYPE:
            raise ValueError(f'Invalid cell_type, should be one of {AVAILABLE_CELL_TYPE}')

        self._cell_type = cell_type

    def set_cell_size_from_curvature(self, n_elements_per_circle: int):

        if type(n_elements_per_circle) is not int:
            raise ValueError('n_elements should be an integer')

        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", n_elements_per_circle)

    def set_meshing_algorithm(self, algorithm: str):
        """Sets the algorithm to use to discretize the domain elements.

        Parameters
        ----------
        algorithm : str
            Name of the tiling algorithm to use.

        """
        if CELL_TYPE[self._cell_type] == 2: # topological_dimension = 2
            if algorithm not in AVAILABLE_MESHING_ALGO_2D.keys():
                raise ValueError(f'Invalid meshing algorithm, can be {list(AVAILABLE_MESHING_ALGO_2D)}')

            gmsh.option.setNumber("Mesh.Algorithm", AVAILABLE_MESHING_ALGO_2D[algorithm])

        elif CELL_TYPE[self._cell_type] == 3: # topological_dimension = 3
            if algorithm not in AVAILABLE_MESHING_ALGO_3D.keys():
                raise ValueError(f'Invalid meshing algorithm, can be {list(AVAILABLE_MESHING_ALGO_3D)}')

            gmsh.option.setNumber("Mesh.Algorithm", AVAILABLE_MESHING_ALGO_3D[algorithm])

        self._algorithm = algorithm


    def set_verbosity(self, verbosity: bool):
        """Sets the verbosity of the considered domain.

        Parameters
        ----------
        verbosity : bool

        """
        if not isinstance(verbosity, bool):
            raise ValueError('verbosity should be a boolean {True, False}')

        if verbosity:
            gmsh.option.setNumber("General.Verbosity", 5)
        else:
            gmsh.option.setNumber("General.Verbosity", 0)

    def discretize(self):
        """Generates a fenics mesh representing the domain.

        Yields
        ------
        mesh : :class:`Mesh<dolfin.cpp.mesh.Mesh>`
            A fenics mesh corresponding to the domain.
        bdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunction>`
            A meshFunction that labels the borders (i.e. outermost facets)
            of the mesh of the domain.
        cdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunction>`
            A meshFunction that labels the inner elements (i.e. cells)
            of the mesh of the domain.
        ds : :class:`Measure<ufl.measure.Measure>`
            An integral measure on the border of the domain of interest.
        dx : :class:`Measure<ufl.measure.Measure>`
            An integral measure within the domain of interest.

        Notes
        -----
        * The method also generates a temporary gmsh mesh
        and stores it temporary on disc.
        * MeshFunctionSizet : `Sizet` means that the output of
        the function is a positive integer.

        """

        if self.mesh is not None:
            self.model.mesh.clear()

        self._mesh_conversion()

        self.ds = fe.Measure('ds', domain=self.mesh, subdomain_data=self.bdata)
        self.dx = fe.Measure('dx', domain=self.mesh, subdomain_data=self.cdata)

    def rotate(self, point, axe, angle):
        dimTags = [(2, val) for val in self.surfaces.values()] + [(3, val) for val in self.volumes.values()]
        self.factory.rotate(dimTags, *point, *axe, angle)
        self.synchronize()

    def _mesh_conversion(self):
        """Converts a temporary gmsh mesh into a fenics (dolphin) one.

        Returns
        -------
        mesh : :class:`Mesh<dolfin.cpp.mesh.Mesh>`
            Fenics mesh corresponding to the domain.
        bdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunction>`
            MeshFunction that labels the borders (i.e. outermost facets)
            of the mesh of the domain.
        cdata : :class:`MeshFunctionSizet<dolfin.cpp.mesh.MeshFunction>`
            MeshFunction that labels the inner elements (i.e. cells)
            of the mesh of the domain.

        """
        # - Get the finite element topological dimension
        ele_dim = CELL_TYPE[self._cell_type]
        log.debug(f"Convert a mesh with topological element of dimension {ele_dim}")

        # - Set default values for the domain & boundaries labels/name
        has_bdata = False  # flag to fill automatically labels to boundary
        has_cdata = False  # flag to fill automatically labels if needed
        sub_boundary_names = {0: 'boundary'}  # default naming for the boundary
        sub_domain_names = {0: 'domain'}  # default naming for the domain


        ### From GMSH to MESHIO
        # Only the rank 0 process generate the gmsh mesh, convert it to MeshIO and save it as XDMF file.
        # It will avoid duplicates
        if MPI_COMM_WORLD.Get_rank() == 0:
            log.debug('Generate GMSH mesh and convert to MeshIO')
            self.model.mesh.generate(ele_dim)

            # - Assert that we will get all the entites of the gmsh mesh during the conversion by catching the number of
            #   entities
            ELEMENT_GMSH_TYPES = {'line': 1, 'triangle': 2, 'tetra': 4} # Gmsh code for a given element type
            element_type, element_tags, _ = self.model.mesh.getElements(ele_dim) # get the entities of dimension `ele_dim`
            N = len(element_tags[np.where(element_type == ELEMENT_GMSH_TYPES[self._cell_type])[0][0]]) # number of entities

            path = self._dir_name + '/mesh.msh'

            gmsh.write(path)
            mesh_io = mio.read(path, file_format='gmsh')

            # - Get the coordinates of the vertices
            ## Note : results if embedded space is != 3 ?
            mesh_io.points = mesh_io.points[:, :self.dim]

            if self._cell_type not in mesh_io.cells_dict:
                raise RuntimeError(f'None elements of the required topological dimension (element_dim = {ele_dim} for '
                                   f'{self._cell_type}) have been found during the conversion from Gmsh to MeshIo.'
                                   f'\nKnown issue 1: no Entities of the required topological dimension ({ele_dim}) '
                                   f'are defined in your Gmsh mesh. Use YourAbstractObject.factory.getEntities(dim={ele_dim}) '
                                   f'to check if it is the case.\nKnown issue 2: some PhysicalGroups have been set for '
                                   f'Entities of topological dimension < {ele_dim}. If PhysicalGroups are defined, be sure '
                                   f'to define them for all the Entities at topological dimension = {ele_dim}. If not, '
                                   f'all the unmarked Entities will be ignored.\nKnown issue 3: ???')
            elif len(mesh_io.cells_dict[self._cell_type]) != N:
                raise Warning(f'Some elements have been lost during the conversion of the gmsh mesh, one possible '
                              f'reason is the partial labelling of entities at topological dimension {ele_dim}. Make '
                              f'sure that you label all the entities ! Unlabelled entities at dim {ele_dim} will be'
                              f' ignored...')

            # - Extract physical properties (and possible name) for the domain
            if ('gmsh:physical' in mesh_io.cell_data_dict.keys()
                    and self._cell_type in mesh_io.cell_data_dict["gmsh:physical"].keys()):
                log.debug(f"Found labels at the dimension {ele_dim}")
                has_cdata = True

                # Notes:
                # 'gmsh: physical' lists the tags given to group of entities of same dimension. Here, the dimension will
                #                  be defined by ele_dim (from the _cell_type).
                #                  There tags can be defined by addPhysicalGroup(ele_dim, ele_tags, tagOfGroup) in gmsh.
                #                  It is possible to set a name for a given PhysicalGroup using setPhysicalName in gmsh.
                #                  Gmsh does not allowed intersection of PhysicalGroup, the first tag for a given entity
                #                  will be the one you will find here.
                #                  Ex: if elements [0, 1] are tagged with 4,
                #                      then 1 is tagged with 2,
                #                      you will see ele. 1 with tag 4 here.
                # Important warning : In the case where some entities have a PhysicialGroup, if it exists entities without
                #                     any PhysicialGroup, they will be ignored !
                #                     TO BE CLEAR : you will loose part of your mesh !
                # Conclusion : using PhysicalGroup, the resulting mesh will always contains PhysicialGroups that form a
                #              partition of the mesh (at the given topological dimension)
                #              However, it is possible that some of the PhysicialGroup do not have a PhysicalName, we will
                #              deal with this situation by setting a default name.

                labels = mesh_io.cell_data_dict["gmsh:physical"][self._cell_type]

                # - Catch PhysicalName associated with PhysicalGroup
                # - reformat field_data as {tag : name} and filter PhysicalName (`ele_dim` topological dimension)
                if mesh_io.field_data:
                    sub_domain_names = {lab_dim[0]: lab_name
                                  for lab_name, lab_dim in mesh_io.field_data.items()
                                  if lab_dim[1] == (ele_dim)}  # tag: name
                    log.debug(f"domains : field data => {sub_domain_names} found")
                else:
                    sub_domain_names = {i: f'sub_domain_{i}' for i in np.unique(labels)}
                    log.debug(f"domains : default field data => {sub_domain_names}")

                if set(labels) != set(sub_domain_names):
                    log.warning('Domain : the number of PhysicalGroup does not correspond to the number of PhysicalName !')
                    log.warning('Domain : some of your PhysicalGroup will be automatically named...')

                    # - set automatic names for the other labels
                    sub_domain_names.update({lab: f'sub_domain_{i}'
                                             for i, lab in enumerate(set(labels) - set(sub_domain_names))})

                mio.write(self._dir_name + '/mesh.xdmf',
                          mio.Mesh(points=mesh_io.points,
                                   cells={self._cell_type:
                                              mesh_io.cells_dict[self._cell_type]},
                                   cell_data={f"{self._cell_type}_labels": [labels]}))
            else:
                log.debug(f"No labels found at the dimension {ele_dim}, use default.")

                mio.write(self._dir_name + '/mesh.xdmf',
                          mio.Mesh(points=mesh_io.points,
                                   cells={self._cell_type:
                                              mesh_io.cells_dict[self._cell_type]}))
            log.debug(f"MeshIO Mesh was written at {self._dir_name + '/mesh.xdmf'}")

            # - Extract physical properties (and possible names) for the boundaries (ele_dim - 1)
            bnd_cell_type = [cell_type for  cell_type, d in CELL_TYPE.items() if d == (ele_dim - 1)][0]

            if ('gmsh:physical' in mesh_io.cell_data_dict.keys()
                    and bnd_cell_type in mesh_io.cell_data_dict["gmsh:physical"].keys()):
                has_bdata = True
                labels = mesh_io.cell_data_dict["gmsh:physical"][bnd_cell_type]

                # - Catch PhysicalName associated with PhysicalGroup
                # - reformat field_data as {tag : name} and filter PhysicalName (`ele_dim` - 1 topological dimension)
                if mesh_io.field_data:
                    sub_boundary_names = {lab_dim[0]: lab_name
                                  for lab_name, lab_dim in mesh_io.field_data.items()
                                  if lab_dim[1] == (ele_dim - 1)}  # tag: name
                    log.debug(f"boundary : field data => {sub_boundary_names} found")
                else:
                    sub_boundary_names = {i: f'sub_boundary_{i}' for i in np.unique(labels)}
                    log.debug(f"boundary : default field data => {sub_boundary_names}")

                if set(labels) != set(sub_boundary_names):
                    log.warning('Boundary: the number of PhysicalGroup does not correspond to the number of PhysicalName !')
                    log.warning('Boundary: some of your PhysicalGroup will be automatically named...')

                    # - set automatic names for the other labels
                    sub_boundary_names.update({lab: f'sub_boundary_{i}'
                                               for i, lab in enumerate(set(labels)-set(sub_boundary_names))})

                bnd_mesh = mio.Mesh(points=mesh_io.points,
                                     cells={bnd_cell_type:
                                              mesh_io.cells_dict[bnd_cell_type]},
                                     cell_data={f"boundary": [labels]})
                mio.xdmf.write(self._dir_name + '/mesh_boundary.xdmf', bnd_mesh)
                log.debug(f"MeshIO Boundary was written at {self._dir_name + '/mesh_boundary.xdmf'}")

        # Update state of other processes from rank 0 (labels and sub-domain names)
        has_cdata = MPI_COMM_WORLD.bcast(has_cdata, root=0)
        has_bdata = MPI_COMM_WORLD.bcast(has_bdata, root=0)
        sub_domain_names = MPI_COMM_WORLD.bcast(sub_domain_names, root=0)
        sub_boundary_names = MPI_COMM_WORLD.bcast(sub_boundary_names, root=0)

        # Synchronize the MPI processes (other processes will wait for the rank 0 for the mesh creation
        MPI_COMM_WORLD.Barrier()

        ### Reading of the XDMF mesh in parallel (all the processes will get the complete mesh)
        log.debug('Generate FeniCS mesh !')
        log.debug(f'{self._cell_type} for {MPI_COMM_WORLD.Get_rank()}')

        # Init FeniCS mesh
        mesh = fe.Mesh()

        # Read the saved mesh
        with fe.XDMFFile(MPI_COMM_WORLD, self._dir_name + '/mesh.xdmf') as f:
            f.read(mesh)

        # Fill the cdata
        if has_cdata:
            mvc = fe.MeshValueCollection("size_t", mesh, ele_dim)
            with fe.XDMFFile(MPI_COMM_WORLD, self._dir_name + '/' + 'mesh.xdmf') as infile:
                infile.read(mvc, f'{self._cell_type}_labels')
            cdata = fe.cpp.mesh.MeshFunctionSizet(mesh, mvc)
        else:
            cdata = _generate_default_cell_data(mesh)

        #  Fill the bdata
        if has_bdata:
            mvc = fe.MeshValueCollection("size_t", mesh, ele_dim - 1)
            with fe.XDMFFile(MPI_COMM_WORLD, self._dir_name + '/mesh_boundary.xdmf') as infile:
                infile.read(mvc, 'boundary')
            bdata = fe.cpp.mesh.MeshFunctionSizet(mesh, mvc)
        else:
            bdata = _generate_default_boundary_data(mesh)

        #  Stored the mesh information
        self.mesh = mesh
        self.cdata = cdata
        self.sub_domain_names = sub_domain_names
        self.bdata = bdata
        self.sub_boundary_names = sub_boundary_names

    def info(self):
        """Prints the descrition of the domain.

        Returns
        -------
        None

        """
        print(self._details)

class GmshProxy:
    def __init__(self, factory):
        self._factory = factory

    def __getattr__(self, name):
        """Redirige les appels de méthode vers factory si le rang est 0."""
        def method(*args, **kwargs):
            #log.debug(f'Call from {MPI_COMM_WORLD.Get_rank()} of the method {name}')
            if MPI_COMM_WORLD.Get_rank() == 0:
                #log.debug(f'Ok for  {MPI_COMM_WORLD.Get_rank()} of the method {name}')
                return getattr(self._factory, name)(*args, **kwargs)
            else:
                log.debug(f'NO for  {MPI_COMM_WORLD.Get_rank()} of the method {method}')
        return method

class BuiltInModel(object):
    """A model based on the Gmsh `geo` factory.

    Attributes
    ----------
    factory : :class:`occ<gmsh.model.geo>`
        A gmsh model `geo` factory to discretize mesh from geometry.

    """

    def __init__(self, *args, **kwargs):
        """Sets the attribute factory to a specific type of gmsh factory.

        Returns
        -------
        factory : :class:`occ<gmsh.model.geo>`
            A gmsh model `geo` factory to discretize mesh from geometry.

        """
        super().__init__()
        self.factory = gmsh.model.geo


class OccModel(object):
    """A model based on the Gmsh `Occ` factory.

    Attributes
    ----------
    factory : :class:`occ<gmsh.model.occ>`
        A gmsh model `occ` factory to discretize mesh from geometry.

    """

    def __init__(self, *args, **kwargs):
        """Sets the attribute factory to a specific type of gmsh factory.

        Returns
        -------
        factory : :class:`occ<gmsh.model.occ>`
            A gmsh model `occ` factory to discretize mesh from geometry.

        """
        super().__init__()
        #self.factory = gmsh.model.occ
        self.factory = GmshProxy(gmsh.model.occ)

    def __sub__(self, other):
        """Cutting the left hand side Domain with the right hand side domain.

        Parameters
        ----------
        other : instance of subclass(AbstractDomain)

        Returns
        -------
        CSGDomain
            Returns th result of the cut operation

        """

        csg = CSGDomain(cell_type=self._cell_type,
                        cell_size=self._cell_size,
                        algorithm=self._algorithm,
                        dim=self.dim)

        self.compute_new_size(other, csg)

        if not self.volumes:
            dim = 2
            objectdimtags = [(dim, tag) for tag in self.surfaces.values()]
            tooldimtags = [(dim, tag) for tag in other.surfaces.values()]
            outdimtag, _ = self.factory.cut(objectdimtags, tooldimtags)
            csg.surfaces[0] = outdimtag[0][1]
        else:
            dim = 3
            objectdimtags = [(dim, tag) for tag in self.volumes.values()]
            tooldimtags = [(dim, tag) for tag in other.volumes.values()]
            outdimtag, _ = self.factory.cut(objectdimtags, tooldimtags)
            csg.volumes[0] = outdimtag[0][1]

        self.factory.synchronize()

        return csg

    def __add__(self, other):
        """Adding the left hand side Domain with the right hand side domain.

        Parameters
        ----------
        other : instance of subclass(AbstractDomain)

        Returns
        -------
        CSGDomain
            Returns th result of the addition operation

        """

        csg = CSGDomain(cell_type=self._cell_type,
                        cell_size=self._cell_size,
                        algorithm=self._algorithm,
                        dim=self.dim)

        self.compute_new_size(other, csg)

        if not self.volumes:
            dim = 2
            objectdimtags = [(dim, tag) for tag in self.surfaces.values()]
            tooldimtags = [(dim, tag) for tag in other.surfaces.values()]
            outdimtag, _ = self.factory.fuse(objectdimtags, tooldimtags)
            csg.surfaces[0] = outdimtag[0][1]
        else:
            dim = 3
            objectdimtags = [(dim, tag) for tag in self.volumes.values()]
            tooldimtags = [(dim, tag) for tag in other.volumes.values()]
            outdimtag, _ = self.factory.fuse(objectdimtags, tooldimtags)
            csg.volumes[0] = outdimtag[0][1]

        self.factory.synchronize()

        return csg

    def __and__(self, other):
        """Compute the intersection of the two domains.

        Parameters
        ----------
        other : instance of subclass(AbstractDomain)

        Returns
        -------
        CSGDomain
            Return the result of the intersection.

        """

        csg = CSGDomain(cell_type=self._cell_type,
                        cell_size=self._cell_size,
                        algorithm=self._algorithm,
                        dim=self.dim)

        self.compute_new_size(other, csg)

        if not self.volumes:
            dim = 2
            objectdimtags = [(dim, tag) for tag in self.surfaces.values()]
            tooldimtags = [(dim, tag) for tag in other.surfaces.values()]
            outdimtag, _ = self.factory.intersect(objectdimtags, tooldimtags)
            csg.surfaces[0] = outdimtag[0][1]
        else:
            dim = 3
            objectdimtags = [(dim, tag) for tag in self.volumes.values()]
            tooldimtags = [(dim, tag) for tag in other.volumes.values()]
            outdimtag, _ = self.factory.intersect(objectdimtags, tooldimtags)
            csg.volumes[0] = outdimtag[0][1]

        self.factory.synchronize()

        return csg

    def compute_new_size(self, other, csg):
        """Computes a characteristic size for the new structure.

        Parameters
        ----------
        other : instance of subclass(AbstractDomain)
            The domain to combine with the considered one.
        csg : `CSGDomain<bvpy.domains.abstract.CSGDomain>`
            The combined domain that must be updated.

        Returns
        -------
        None

        """

        all_sizes = list(self._characteristic_size.values())\
            + list(other._characteristic_size.values())
        csg._characteristic_size = {'width': max(all_sizes)}


class CSGDomain(AbstractDomain, OccModel):
    def __init__(self, *args, **kwargs):
        """The resulting class of an add/sub or intersection of Domains.

        """
        super(CSGDomain, self).__init__(**kwargs)

    def geometry(self, *args, **kwargs):
        pass
